import numpy as np
import pandas as pd


class Phenotype:
    """
    Phenotype class
    
    Examples
    --------
    >>> ph = Phenotype(D=10, capacity=1000)
    >>> ph.add_center(0)
    >>> ph._used_centers
    >>> ph.replace(0,8,11.9)
    >>> ph.L()
    >>> ph.replace(0,2,11.9)
    >>> ph.add_center(1)
    >>> ph.add_center(2)
    >>> ph.replace(1,2,13.9)
    >>> ph.replace(2,2,15.9)
    >>> ph.to_pandas()
         0    1     2    3    4    5    6    7     8    9
    0  0.0  0.0  11.9  0.0  0.0  0.0  0.0  0.0  11.9  0.0
    1  0.0  0.0  13.9  0.0  0.0  0.0  0.0  0.0   0.0  0.0
    2  0.0  0.0  15.9  0.0  0.0  0.0  0.0  0.0   0.0  0.0
    >>> ph.delete_center(1)
    >>> ph.to_pandas()
         0    1     2    3    4    5    6    7     8    9
    0  0.0  0.0  11.9  0.0  0.0  0.0  0.0  0.0  11.9  0.0
    2  0.0  0.0  15.9  0.0  0.0  0.0  0.0  0.0   0.0  0.0
    >>> ph.reverse_changes()
    Empty DataFrame
    Columns: [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    Index: []
    """
    def __init__(self, D, capacity):
        self._possible_centers = np.zeros((capacity, D), dtype=np.float)
        self._capacity = capacity
        self._free_centers = list(range(capacity))[::-1]
        self._used_centers = []
        self._changes_tracker = []
        self._centers_creation = []

    def replace(self, c, d, x, memo=True):
        old_x = self._possible_centers[c,d]
        self._possible_centers[c,d] = x
        if memo: self._changes_tracker.append((c,d,old_x))

    def reverse_changes(self):
        while self._changes_tracker:
            c,d,x = self._changes_tracker.pop()
            self.replace(c, d, x, False)
        while self._centers_creation:
            c,a = self._centers_creation.pop()
            if a > 0:
                self.delete_center(c,False)
            if a < 0:
                self.add_center(c,False)

    def delete_center(self,c,memo=True):
        self._used_centers.remove(c)
        self._free_centers.append(c)
        if memo: self._centers_creation.append((c, -1))

    def empty_center_candidate(self):
        return(self._free_centers[0])

    def add_center(self,c,memo=True):
        self._free_centers.remove(c)
        self._used_centers.append(c)
        if memo: self._centers_creation.append((c, 1))

    def apply_changes(self):
        self._changes_tracker = []
        self._centers_creation = []

    def L(self):
        return(self._possible_centers[self._used_centers,:])

    def to_pandas(self):
        L = pd.DataFrame(self.L(), index = self._used_centers)
        return(L)



import numpy as np
import pandas as pd

class Genotype:
    """
    Genotype class

    Examples
    --------
    >>> np.random.seed(0)
    >>> g = Genotype()
    >>> g.insertion((3,4,9))
    >>> g.insertion((3,4,99))
    >>> g.nb_centers
    1
    >>> g.insertion((3,40,7))
    >>> g.insertion((1,40,90))
    >>> g.genes
    {(1, 40, 3), (3, 4, 0), (3, 4, 1), (3, 40, 2)}
    >>> g.to_pandas()
        4    40
    3  2.0  1.0
    1  NaN  1.0
    >>> g.nb_centers
    2
    >>> g.G
    4
    >>> candidate = g.get_gene_candidate()
    >>> candidate
    (3, 40, 2)
    >>> g.deletion(candidate)
    >>> g.G
    3
    >>> g.genes
    {(1, 40, 3), (3, 4, 0), (3, 4, 1)}
    >>> g.to_pandas()
        4    40
    3  2.0  NaN
    1  NaN  1.0
    >>> g.deletion(g.get_gene_candidate())
    >>> g.deletion(g.get_gene_candidate())
    >>> g.deletion(g.get_gene_candidate())
    Empty DataFrame
    Columns: []
    Index: []
    """
    def __init__(self):
        self.genes = set()
        self.candidate_del = None
        self._t = 0
        self.genes_counter = {}
        self.center_counter = {}
        self.G = 0
        self.nb_centers = 0

    def deletion(self, gene):
        """
        Delete a gene from the genome

        Parameters
        ----------
        gene : tuple
            Tuple containing 3 elements: the cluster id, the feature id and an id
        """
        c,d,_ = gene
        self.genes.remove(gene)
        self.genes_counter[c][d] -= 1
        self.center_counter[c] -= 1
        self.G -= 1
        if not self.genes_counter[c][d]:
            self.genes_counter[c].pop(d)
        if not self.center_counter[c]:
            self.nb_centers -= 1
            self.center_counter.pop(c)
            self.genes_counter.pop(c)

    def insertion(self,gene):
        """
        Insert a new gene into the genome

        Parameters
        ----------
        gene : tuple
            Tuple containing 3 elements: the cluster id, the feature id and a
            coordinate
        """
        c,d,_ = gene
        self.genes.add((c,d,self._t,))
        if c not in self.center_counter:
            self.center_counter[c] = 0
            self.genes_counter[c] = {}
            self.nb_centers += 1
        if d not in self.genes_counter[c]:
            self.genes_counter[c][d] = 0
        self.center_counter[c] += 1
        self.genes_counter[c][d] += 1
        self.G += 1
        self._t += 1

    def get_gene_candidate(self):
        """
        Return a random gene from the genome
        """
        candidate = self.genes.pop()
        self.genes.add(candidate)
        return(candidate)

    def to_pandas(self,index=None,columns=None):
        """
        Convert the genome in a pandas data frame

        Parameters
        ----------
        index : numpy.ndarray or list, default=None
            list containing the index name
        columns : numpy.ndarray or list, default=None
            list containing the columns name

        Returns
        -------
        pandas.DataFrame
            Data Frame containing the genome information: index represent
            centers, while columns denote features
        """
        return pd.DataFrame(self.genes_counter,index,columns).T



import numpy as np
import pandas as pd
from sklearn.metrics import pairwise_distances
from tqdm.autonotebook import tqdm
from SubCMedians.model import Model

class subcmedians_v1:
    def __init__(self, D, Gmax=300, H=200, nb_iter=10000,random_state=None):
        """SubCMedians subspace clustering.
        Parameters
        ----------
        Gmax : int, default=300
            Maximal model size, i.e., maximal number of coordinates that can
            be included in the model
        nb_iter : int, default=10000
            Number of iterations of the SubCMedians algorithm
        random_state : int, RandomState instance, default=None
            Determines random number generation for centroid initialization. Use
            an int to make the randomness deterministic.
            See :term:`Glossary <random_state>`.
        Attributes
        ----------
        cluster_centers_ : ndarray of shape (n_clusters, n_features)
            Coordinates of cluster centers.
        subspaces_  : ndarray of shape (n_clusters, n_features)
            Subspace of cluster centers.
        labels_ : ndarray of shape (n_samples,)
            Labels of each point
        sae_ : float
            Sum of Absolute Errors between data instances and their closest center

        Examples
        --------
        >>> from SubCMedians.subcmedians import subcmedians
        >>> from SubCMedians.data_generator import make_subspace_blobs
        >>> D = 20
        >>> dataset_params={"p_dim": 0.7,
                            "n_samples":5000,
                            "n_features":D,
                            "centers":12}
        >>> X,y_true,ss = make_subspace_blobs(**dataset_params)
        >>> X = (X - X.mean(axis=0))/ X.std(axis=0)
        >>> scm = subcmedians(D, Gmax=300, H=200, nb_iter=10000)
        >>> scm.fit(X)
        >>> scm.predict(X)
        """
        self.model = Model(D=D,capacity=Gmax)
        self.Gmax = Gmax
        self.H = H
        self.D = D
        self.nb_iter = nb_iter
        self.S = {}
        self.random_state = random_state

    def predict(self, X):
        """
        Compute the cluster membership of the data instance from $X$

        Parameters
        ----------
        X : numpy.array
            Data set to be clustered, rows represent instances (points) and
            columns represent features (dimensions)

        Returns
        -------
        numpy.array
            cluster memberships

        """
        if len(X.shape) == 1:
            X = X.reshape(1,-1)
        distances = pairwise_distances(X,
                                       self.model.pheno.L(),
                                       metric="l1",
                                       n_jobs=None)
        self.labels_ = np.argmin(distances,axis=1)
        self.sae_ = distances[range(distances.shape[0]),self.labels_]
        return(self.labels_)

    def sae_score(self, X):
        """
        Compute the Sum of Absolute Errors (SAE) of dataset $X$ with respect to
        the current clustering model, a low SAE means a better quality

        Parameters
        ----------
        X : numpy.array
            Data set to be clustered, rows represent instances (points) and
            columns represent features (dimensions)

        Returns
        -------
        float
            Sum of Absolute Errors

        """
        if len(X.shape) == 1:
            X = X.reshape(1,-1)
        distances = pairwise_distances(X,
                                       self.model.pheno.L(),
                                       metric="l1",
                                       n_jobs=None)
        sae = distances.min(axis=1).sum()
        return(sae)

    def _predict_point_center(self, point):
        distances = pairwise_distances(point.reshape(1,-1),
                                       self.model.pheno.L(),
                                       metric="l1",
                                       n_jobs=None)
        distances = distances.flatten()
        y_index = np.argmin(distances)
        ae = distances[y_index]
        id_c = self.model.pheno._used_centers[y_index]
        center = self.model.pheno._possible_centers[id_c,:]
        distances_dimensions = np.abs(center - point)
        return(id_c, ae, distances_dimensions)

    def _choose_insertion(self, point):
        if np.random.random() > self.model.geno.G/(self.model.geno.G+1):
            c = self.model.pheno.empty_center_candidate()
            distances_dimensions = np.abs(point)
        else:
            c, ae, distances_dimensions = self._predict_point_center(point)
        # Choose insertion dim
        #d = np.argmax(distances_dimensions)#np.random.randint(self.D)#
        d = np.random.randint(self.D)
        x = point[d]
        return(c,d,x)

    def _choose_deletion(self, point):
        candidate = self.model.geno.get_gene_candidate()
        return(candidate)

    def _insertion(self,point):
        self.model._candidate_insertion = self._choose_insertion(point)
        self.model.try_insertion()

    def _deletion(self,point):
        self.model._candidate_deletion = self._choose_deletion(point)
        self.model.try_deletion()

    def _model_candidate(self, point):
            if self.model.geno.G > self.Gmax:
                self._deletion(point)
            self._insertion(point)


    def fit(self, X, lazy=False):
        """
        Build the SubCMedians model associated to dataset $X$

        Parameters
        ----------
        X : numpy.array
            Data set to be clustered, rows represent instances (points) and
            columns represent features (dimensions)

        Returns
        -------
        subcmedians
            Fitted subcmedians instance

        """
        if self.random_state is not None:
            np.random.seed(self.random_state)
        industrius = (not lazy)
        if X.shape[0] < self.H:
            self.H = X.shape[0]-1
        self.S = X[:self.H,:].copy()
        self._sae_history = []
        self._nb_centers_history = []
        self._genome_size_history = []
        #self.genome_model = []
        #self.pheno_model = []
        #self.changes_accepted = []
        #self.changes = []
        i = self.H
        h = 0
        self._model_candidate(self.S[h])
        sae = self.sae_score(self.S)
        for t in tqdm(range(self.nb_iter)):
            # update dataset and SAE
            ae_old_point = self.sae_score(self.S[h,:])
            point = X[i,:]
            self.S[h,:] = point
            ae_new_point = self.sae_score(self.S[h,:])
            sae = sae - ae_old_point + ae_new_point
            # generate candidate
            if industrius or ae_old_point < ae_new_point:
                self._model_candidate(point)
            #self.changes.append([self.model._candidate_deletion,self.model._candidate_insertion])
            sae_ = self.sae_score(self.S)
            if sae >= sae_:
                sae = sae_
                self.model.apply_changes()
                #self.changes_accepted.append(1)
            else:
                self.model.reverse_changes()
                #self.changes_accepted.append(0)
            self._sae_history.append(sae)
            self._nb_centers_history.append(self.model.geno.nb_centers)
            self._genome_size_history.append(self.model.geno.G)
            #self.genome_model.append(self.model.geno.to_pandas())
            #self.pheno_model.append(self.model.pheno.to_pandas())
            h = (h+1)%self.H
            i = (i+1)%X.shape[0]
        self.cluster_centers_ = self.model.pheno.to_pandas()
        self.subspaces_ = self.model.pheno.to_pandas()*0
        self.subspaces_ += self.model.geno.to_pandas()
        self.subspaces_ = self.subspaces_[self.subspaces_.any(axis=1)]
        self.sae_ = sae
        return(self)



import numpy as np
import pandas as pd
from SubCMedians.phenotype import Phenotype
from SubCMedians.genotype import Genotype

class Model:
    """
    Model Class

    Examples
    --------
    >>> m = Model(10,100)
    >>> m._candidate_insertion = (0,1,200)
    >>> m.try_insertion()
    >>> m.pheno.to_pandas()
    >>> m.geno.to_pandas()
    >>> m.apply_changes()
    >>> m._candidate_insertion = (1,9,20)
    >>> m.try_insertion()
    >>> m.apply_changes()
    >>> m._candidate_insertion = (1,9,8990)
    >>> m.try_insertion()
    >>> m.pheno.to_pandas()
    >>> m.reverse_changes()
    >>> m.pheno.to_pandas()
    >>> m._candidate_deletion = m.geno.get_gene_candidate()
    >>> m.try_deletion()
    >>> m.apply_changes()
    >>> m.pheno.to_pandas()
    >>> m.geno.to_pandas()
    >>> m._candidate_insertion = (0,1,20)
    >>> m.try_insertion()
    >>> m.apply_changes()
    >>> m.pheno.to_pandas()
    >>> m.geno.to_pandas()
    >>> m._candidate_deletion = m.geno.get_gene_candidate()
    >>> m._candidate_insertion = m._candidate_deletion[:2]+(666,)
    >>> m.try_deletion()
    >>> m.try_insertion()
    >>> m.apply_changes()
    >>> m.pheno.to_pandas()
    >>> m._candidate_insertion = (3,2,666,)
    >>> m.try_insertion()
    >>> m.apply_changes()
    >>> m.pheno.to_pandas()
    >>> m.geno.to_pandas()
    >>> m._candidate_deletion = m.geno.get_gene_candidate()
    >>> m.try_deletion()
    >>> m.pheno.to_pandas()
    >>> m.geno.to_pandas()
    >>> m._candidate_insertion = m._candidate_deletion[:2]+(999,)
    >>> m.try_insertion()
    >>> m.pheno.to_pandas()
    >>> m.geno.to_pandas()
    >>> m._candidate_deletion = (3, 2, 5)
    >>> m.try_deletion()
    >>> m.pheno.to_pandas()
    >>> m.geno.to_pandas()
    >>> m._candidate_insertion = m._candidate_deletion[:1]+(4,999,)
    >>> m.try_insertion()
    >>> m.apply_changes()

    """
    def __init__(self, D, capacity):
        self.pheno = Phenotype(D,capacity)
        self.geno = Genotype()
        self._candidate_deletion = None
        self._candidate_insertion = None

    def try_deletion(self):
        if self._candidate_deletion is not None:
            c,d,_ = self._candidate_deletion
            if self.geno.genes_counter[c][d] == 1:
                self.pheno.replace(c,d,0)
            if self.geno.center_counter[c] == 1:
                self.pheno.delete_center(c)

    def try_insertion(self):
        if self._candidate_insertion is not None:
            c,d,x = self._candidate_insertion
            self.pheno.replace(c,d,x)
            if c not in self.pheno._used_centers:
                self.pheno.add_center(c)

    def reverse_changes(self):
        self.pheno.reverse_changes()
        self._candidate_deletion = None
        self._candidate_insertion = None

    def apply_changes(self):
        self.pheno.apply_changes()
        if self._candidate_deletion is not None:
            self.geno.deletion(self._candidate_deletion)
        if self._candidate_insertion is not None:
            self.geno.insertion(self._candidate_insertion)
        self._candidate_deletion = None
        self._candidate_insertion = None



import numpy as np
from sklearn.datasets import make_blobs

def make_subspace_blobs(p_dim=0.7,**blobs_params):
    """
    Parameters
    ----------
    p_dim : float, default=0.7
        Proportion of noise dimensions in clusters
    **blobs_params : named parameters
        Parameters from the sklearn.datasets.blobs_params function

    Returns
    -------
    np.ndarray
        Dataset generated
    np.ndarray
        cluster memberships
    np.ndarray
        True cluster subspaces

    Examples:
    ---------
        >>> import seaborn as sns
        >>> import matplotlib.pyplot as plt
        >>> import pandas as pd
        >>> X,y_true,subspaces = make_subspace_blobs()
        >>> X_df = pd.DataFrame(X)
        >>> X_df["cluster"] = y_true
        >>> sns.scatterplot(x=0, y=1, hue="cluster", marker=".", data=X_df)
        >>> plt.show()
        >>> sns.heatmap(subspaces)
        >>> plt.show()
    """
    default_blobs_params = {"n_samples":1000,
                            "n_features":10,
                            "centers":5,
                            "cluster_std":1.0,
                            "center_box":(-10.0, 10.0),
                            "shuffle":True,
                            "random_state":None,}
    default_blobs_params.update(blobs_params)
    X,y_true = make_blobs(**blobs_params)
    subspaces = np.random.binomial(size=(default_blobs_params["centers"],
                                         default_blobs_params["n_features"]),
                                   p=1 - p_dim,
                                   n=1)
    uniform_coordinates = np.random.uniform(low=default_blobs_params["center_box"][0],
                                            high=default_blobs_params["center_box"][1],
                                            size=X.shape)
    X_rand_subspaces =  subspaces[y_true,:]
    X *= np.logical_not(X_rand_subspaces)
    X += uniform_coordinates*X_rand_subspaces
    return(X,y_true,np.logical_not(subspaces))



from scipy.optimize import linear_sum_assignment
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
import numpy as np
epsilon = 1e-5

def plot_projection(centers,X,y_pred,x=0,y=1):
    X_full = pd.DataFrame(X)
    X_full["cluster"] =  ["clus. "+str(v) for v in y_pred]
    X_full["cluster"] = X_full["cluster"].astype('category')
    sns.scatterplot(x=x, y=y, hue='cluster', marker=".", data=X_full)
    sns.scatterplot(x=x, y=y, marker="o", data=pd.DataFrame(centers))
    plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
    plt.show()

def plot_evolution(variables,name):
    for k in variables:
        plt.plot(variables[k],label=k)
    plt.xlabel("Iterations",fontsize=20)
    plt.ylabel(name,fontsize=20)
    plt.legend()
    plt.tight_layout()
    plt.show()

def cross_tab(y_pred,y_true):
    res = pd.DataFrame([y_true,y_pred]).T
    return(pd.crosstab(y_true,y_pred))

def plot_cross_tab(y_pred,y_true):
    ct = cross_tab(y_pred,y_true)
    sns.heatmap(ct,annot=True)
    b, t = plt.ylim()
    b += 0.5
    t -= 0.5 
    plt.ylim(b, t)
    plt.tight_layout()

def _mapped(crosstab):
    mapped_clusters = (crosstab.T * 1. / crosstab.sum(1)).T
    return mapped_clusters == mapped_clusters.max(0)

def entropy_score(crosstab):
    found_clusters_effective = crosstab.sum(0)
    p_h_in_c = crosstab * 1. / (found_clusters_effective + epsilon)
    log_p_h_in_c = np.log(p_h_in_c+epsilon)
    pre_ec = -1. * p_h_in_c * log_p_h_in_c
    pre_ec = pre_ec.fillna(0)
    ec = pre_ec.sum(0)
    num = (ec * found_clusters_effective).sum()
    denum = found_clusters_effective.sum() * np.log(len(crosstab.index))
    return 1. - num * 1. / denum

def accurracy_score(crosstab):
    best_matching_y_true = crosstab == crosstab.max(0)
    best_matching_y_true_weight = 1./best_matching_y_true.sum(0)
    correctly_predicted_objects = crosstab * best_matching_y_true * best_matching_y_true_weight
    return correctly_predicted_objects.sum().sum() * 1. / crosstab.sum().sum()

def ce_score(crosstab):
    cost = crosstab.values
    row_ind, col_ind = linear_sum_assignment(-cost)
    best_assignment = cost[row_ind, col_ind].sum()
    return(best_assignment / (cost.sum().sum() + epsilon))

def recall_score(crosstab):
    mapped_clusters = _mapped(crosstab)
    mapped_clusters = (crosstab.T * 1. / crosstab.sum(1)).T
    num = mapped_clusters * crosstab
    num = num.sum(1)
    denum = crosstab.sum(1)
    return num * 1. / (denum + epsilon)

def precision_score(crosstab):
    mapped_clusters = _mapped(crosstab)
    num = mapped_clusters * crosstab
    num = num.sum(1)
    denum = mapped_clusters * crosstab.sum(0)
    denum = denum.sum(1)
    return num * 1. / (denum + epsilon)

def f1_score(crosstab):
    recall = recall_score(crosstab)
    precision = precision_score(crosstab)
    denum = recall + precision
    num = 2 * recall * precision
    return(num/denum)

def evaluate(y_true,y_pred):
    crosstab = cross_tab(y_true,y_pred).T
    entropy = entropy_score(crosstab)
    accuracy = accurracy_score(crosstab)
    ce = ce_score(crosstab)
    recall = recall_score(crosstab)
    precision = precision_score(crosstab)
    f1 = f1_score(crosstab)
    results={"accuracy":accuracy,
             "f1":f1.mean(),
             "precision":precision.mean(),
             "recall":recall.mean(),
             "entropy":entropy,
             "ce":ce}
    return(results)
